#!/bin/bash

DEPENDS_FILE=depends/depends.txt
RULES_FILE=depends/rules.txt
BUILD_DIR="$1"
ROOT_DIR="$PWD"
NPROC=`getconf _NPROCESSORS_ONLN`
ABS_BUILD_DIR=`realpath $BUILD_DIR`

function die ()
{
    echo "ERROR: $*"
    exit 1
}

function curl ()
{
    $CURL $1 --output $2 2>/dev/null >/dev/null
    [ $? -ne 0 ] && die "Failed to download $1 and save it to $2"
}

function wget ()
{
    $WGET $1 -O $2 2>/dev/null >/dev/null
    [ $? -ne 0 ] && die "Failed to download $1"
}

function git ()
{
    local URL=$1
    local DEST=$2
    shift 2
    $GIT clone $URL $DEST $*
    [ $? -ne 0 ] && die "Failed to clone $URL into folder $DEST with options $OPT"
}

function cmake ()
{
    $CMAKE $*
}

function get ()
{
    local NUM=$1
    shift
    echo $* | cut -d' ' -f$NUM -
}

function get_from ()
{
    local NUM=$1
    shift
    echo $* | cut -d' ' -f$NUM- -
}

function untar ()
{
    local OLD=`tar tf depends/$1 | cut -d'/' -f1 - | uniq`
    (cd depends && $TAR xzf $1 --transform="s!^$OLD!$2!")
}

[ "x${BUILD_DIR}" = "x" ] && die "You must specify the lektor's build folder as the first argument"
echo "Lektor's build file is $BUILD_DIR"

CURL=`which curl 2>/dev/null`
[ $? -ne 0 ] && die "Failed to find the curl binary"
WGET=`which wget 2>/dev/null`
[ $? -ne 0 ] && die "Failed to find the wget binary"

TAR=`which tar 2>/dev/null`
[ $? -ne 0 ] && die "Failed to find the tar binary"
GIT=`which git 2>/dev/null`
[ $? -ne 0 ] && die "Failed to find the git binary"
CMAKE=`which cmake 2>/dev/null`
[ $? -ne 0 ] && die "Failed to find the cmake binary"

! [ -f $DEPENDS_FILE ] && die "You must run this script from the root of the project"
! [ -f $PWD/scripts/depends.bash ] && die "You must run thus script from the root of the project"
! [ -d $BUILD_DIR/depends_install ] && mkdir $BUILD_DIR/depends_install
! [ -d $BUILD_DIR/depends_install ] && die "Failed to create depends install folder"

#########################
# Download dependencies #
#########################

! [ -d depends ] && mkdir depends

while read LINE
do
    FUNC=`get 1 $LINE`
    FOLD=`get 2 $LINE`
    URL=`get 3 $LINE`
    DEST=`get 4 $LINE`
    OPT=`get_from 5 $LINE`

    if [ -f depends/$DEST ] || [ -d depends/$DEST ]
    then
        echo "Dependency $FOLD is already downloaded," \
             "remove depends/$DEST and depends/$DEST to re-download it"
        continue
    fi

    echo "Downloading $FOLD from $URL to depends/$DEST"
    case $FUNC in
        wget)
            wget $URL depends/$DEST
            ;;
        curl)
            curl $URL depends/$DEST
            ;;
        git)
            git $URL depends/$DEST $OPT
            ;;
    esac
done < $DEPENDS_FILE

##################################
# Extract dependencies and build #
##################################

while read LINE
do
    FOLD=`get 2 $LINE`
    FILE=`get 4 $LINE`
    if [ -d depends/$FOLD ]
    then
        echo "Skip already extracted folder depends/$FOLD"
        continue
    fi
    echo "Extract file depends/$FILE to depends/$FOLD"
    untar $FILE $FOLD
done < $DEPENDS_FILE

while read LINE
do
    FOLD=`get 1 $LINE`
    TYPE=`get 2 $LINE`
    OPT=`get_from 3 $LINE`

    # Replace __VARIABLES__ from the OPT variable
    OPT=`echo "$OPT" | sed -e "s+__BUILD_DIR__+$BUILD_DIR/depends_install/+g"`

    ! [ -d $BUILD_DIR/depends/$FOLD ] && mkdir -p $BUILD_DIR/depends/$FOLD
    ! [ -d $BUILD_DIR/depends/$FOLD ] && die "Failed to create the build folder" \
                                             "for $FOLD"

    if [ -f $BUILD_DIR/depends/$FOLD/.lektor ]
    then
        echo "Skip already built folder $FOLD, if you want to rebuild it," \
             "remove file $BUILD_DIR/depends/$FOLD/.lektor"
        continue
    fi

    # Go to right folder
    cd $BUILD_DIR/depends/$FOLD

    # Install method, i.e. install_sw or install
    INSTALL="install"

    # Types of builds
    case "$TYPE" in
        configure)
            $ROOT_DIR/depends/$FOLD/configure --prefix=$ABS_BUILD_DIR/depends_install CC=$CC CXX=$CXX $OPT
            [ $? -ne 0 ] && die "Failed to configure $FOLD"
            ;;
        zconfigure)
            # The CC= and CXX= doesn't work for zlib
            # TODO pour le Kubat du futur: trouver un moyen uniforme de passer le '-fPIC'
            CC=$CC CXX=$CXX CFLAGS="-fPIC -Wno-error" $ROOT_DIR/depends/$FOLD/configure --prefix=$ABS_BUILD_DIR/depends_install $OPT
            [ $? -ne 0 ] && die "Failed to configure $FOLD"
            ;;
        sslconfigure)
            # Yes, we need to call Configure and not configure here...
            # TODO pour le Kubat du futur: trouver un moyen de pas ardcoder le linux-x86_64 (build FreeBSD, OSX et autres)
            CC=$CC CXX=$CXX $ROOT_DIR/depends/$FOLD/Configure --prefix=$ABS_BUILD_DIR/depends_install $OPT linux-x86_64
            [ $? -ne 0 ] && die "Failed to configure $FOLD"
            INSTALL="install_sw"
            ;;
        cmake)
            # Yes, some projects build with cmake, will be deleted in future commits
            CC=$CC CXX=$CXX cmake $ROOT_DIR/depends/$FOLD/ $OPT
            [ $? -ne 0 ] && die "Failed to configure $FOLD"
    esac

    # Common part
    make -j$NPROC
    [ $? -ne 0 ] && die "Failed to make $FOLD, build folder is $BUILD_DIR/depends/$FOLD"
    make $INSTALL
    [ $? -ne 0 ] && die "Failed to install $FOLD, build folder is " \
                        "$BUILD_DIR/depends/$FOLD, install folder " \
                        "is $ROOT_DIR/depends_install"

    # Exit folder and mark it as compiled
    touch .lektor
    cd $ROOT_DIR
    echo "Finished to build $FOLD, result is in $BUILD_DIR/depends/$FOLD"
done < $RULES_FILE

# vim:syntax=sh ts=4 sw=4 expandtab
