#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/database.h>

#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <sqlite3.h>

#include "macro.h"

bool
database_config_set(volatile sqlite3 *db, const char *section, const char *key, const char *value)
{
    static const char *SQL_STMT =
        "INSERT OR REPLACE INTO"
        " config (section, key, value)"
        " VALUES (?, ?, ?);\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    int code;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_BIND_TEXT(db, stmt, 3, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK && code != SQLITE_DONE) {
        LOG_ERROR("DB", "Failed to insert or replace: %s",
                  sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    LOG_INFO_CONT("Set [%s.%s] to %s", section, key, value);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_text(volatile sqlite3 *db, const char *section, const char *key, char *value,
                         size_t len)
{
    static const char *SQL_STMT =
        "SELECT value"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    char *row;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    row = (char *) sqlite3_column_text(stmt, 0);
    safe_strncpy(value, row, len);
    value[len - 1] = 0;
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_text_nospace(volatile sqlite3 *db, const char *section, const char *key,
                                 char *value, size_t len)
{
    static const char *SQL_STMT =
        "SELECT REPLACE(value, ' ', '')"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    char *row;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    row = (char *) sqlite3_column_text(stmt, 0);
    strncpy(value, row, len);
    value[len - 1] = 0;
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_exists(volatile sqlite3 *db, const char *section, const char *key)
{
    static const char *SQL_STMT =
        "SELECT value"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_int(volatile sqlite3 *db, const char *section, const char *key, int *value)
{
    static const char *SQL_STMT =
        "SELECT CAST(value AS INTEGER)"
        " FROM config"
        " WHERE section COLLATE nocase = ? AND key COLLATE nocase = ?;";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);

    SQLITE_STEP_ROW(db, stmt, error);

    *value = sqlite3_column_int(stmt, 0);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_queue(volatile sqlite3 *db, const char *option, int value)
{
    static const char *SQL_STMT_TMP = "UPDATE queue_state SET %s = ? WHERE id = 42;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    int code;

    if (value < 0)
        value = 0;
    if (value > 100)
        value = 100;

    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TMP, option);
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK || code != SQLITE_ROW || code != SQLITE_DONE)
        goto error;

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_validate_conf(volatile sqlite3 *db)
{
    const char *section;
#define section(_sct)   section = _sct;
#define value_opt(name, value)
#define value(name, value)                                                    \
    if (!database_config_exists(db, section, name)) {                         \
        LOG_ERROR("CONFIG", "Missing option \""name"\" in section \"%s\"",\
                      section);                                               \
        return false;                                                         \
    }
#include <lektor/config.def>
#undef section
#undef value
#undef value_opt
    return true;
}

bool
database_config_dump(volatile sqlite3 *db, FILE *f)
{
    static const char *SQL = "SELECT section, key, value FROM config;";
    sqlite3_stmt *stmt = 0;

    SQLITE_PREPARE(db, stmt, SQL, error);

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        if (sqlite3_column_type(stmt, 0) == SQLITE_NULL) {
            fprintf(f, "NO SECTION => %s => %s\n",
                    sqlite3_column_chars(stmt, 1),
                    sqlite3_column_chars(stmt, 2));
        } else {
            fprintf(f, "%s => %s => %s\n",
                    sqlite3_column_chars(stmt, 0),
                    sqlite3_column_chars(stmt, 1),
                    sqlite3_column_chars(stmt, 2));
        }
    }

    sqlite3_finalize(stmt);
    return true;
error:
    sqlite3_finalize(stmt);
    return false;
}

bool
database_config_obfuscation_check(volatile sqlite3 *db, int obfuscation)
{
    static const char *GET_SQL = "SELECT obfuscation FROM misc WHERE id = 42;";
    static const char *SET_SQL = "UPDATE misc SET obfuscation = ? WHERE id = 42;";
    sqlite3_stmt *stmt = 0;

    SQLITE_PREPARE(db, stmt, GET_SQL, error);

    if (SQLITE_ROW == sqlite3_step(stmt)) {
        /* Obfuscation mode set */
        bool ret = obfuscation == sqlite3_column_int(stmt, 0);
        sqlite3_finalize(stmt);
        return ret;
    }

    else {
        /* No obfuscation mode set */
        sqlite3_finalize(stmt);
        SQLITE_PREPARE(db, stmt, SET_SQL, error);
        SQLITE_BIND_INT(db, stmt, 1, obfuscation, error);
        SQLITE_STEP_DONE(db, stmt, error);
        sqlite3_finalize(stmt);
        return true;
    }

error:
    sqlite3_finalize(stmt);
    return false;
}
